//=============================================================================
// File Name: osboost.cc
// Written by: Shang-Tse Chen (27.05.12)
//=============================================================================
#include <numeric>
#include <math.h>
#include "osboost.h"

//==============================================//
//-------------- Tool Functions ----------------//
//==============================================//
int choose(double*p, int n)
{
    double r = (rand() / (double)(RAND_MAX));
    double norm = 0.0;
    for(int i=0; i<n; i++)
        norm += p[i];
    for(int i=0; i<n; i++)
        p[i] /= norm; 
    double cdf = 0.0;
    for(int i=0; i<n; i++){
        cdf += p[i];
        if(r<cdf)
            return i+1;
    }
    return n;
}

bool mycomp (double i,double j) { return (i>j);}
bool is_negative(double a) { return (a<0.0);}

void project_to_simplex(double* v, int n, double b)
{
    double u[n];
    double sv[n];
    int rho;
    for(int i=0; i< n; i++){
        v[i] = v[i]>0? v[i] : 0.0;
        u[i] = v[i];
    }
    std::sort(u, u+n, mycomp);
    std::partial_sum(u, u+n, sv);
    for(int i=0; i<n; i++)
        u[i] -= (sv[i]-b)/(double)(i+1);
    double* it = std::find_if(u, u+n, is_negative); 
    rho = it - u - 1;
    double theta = std::max(0.0, (sv[rho] - b) / (rho+1.0));
    for(int i=0; i< n; i++){
        v[i] = std::max(v[i] - theta, 0.0);
    }
}

void L2_normalize(double* x, int n, double b)
{
    double norm = 0.0;
    for(int i=0; i< n; i++)
        norm += x[i] * x[i];
    norm = sqrt(norm);
    if(norm > b){
        for(int i=0; i< n; i++)
            x[i] /= norm;
    }
}

//==============================================//
//---------------- Learner ---------------------//
//==============================================//
int Learner::predict(simple_sparse_vector& x)
{
    return (raw_predict(x) > 0.0)? 1 : -1;
}

//==============================================//
//-------------- Perceptron --------------------//
//==============================================//
void Perceptron::reset()
{
    t = 0;
    w.reset();
}

double Perceptron::raw_predict(simple_sparse_vector& x)
{
    return (w * x); 
}

double Perceptron::normalized_predict(simple_sparse_vector& x)
{
    if ( w.snorm() > 1.0)
        return (w * x) / sqrt(w.snorm()); 
    else
        return (w * x);
}

void Perceptron::update(Example& ex)
{
    ++t;
    if(raw_predict(ex.x) * ex.y <= 0.0){
        double eta = 1.0;  
        w.add(ex.x, eta * ex.w * ex.y);
    }
}

//==============================================//
//---------- Naive Bayes (Gaussian)-------------//
//==============================================//
void NaiveBayesGauss::update(Example& ex)
{
    int label = (ex.y == -1)? 0 : 1;
    double old_sum_w = sum_w;
    sum_w += ex.w;
    ++t;

    // update P( C )
    pc[label] += ex.w;

    // update P( Fi | C )
    // calculate variance by the method of Donald E. Knuth (1998).
    // The Art of Computer Programming, volume 2, 3rd edn., p. 232.
    for(int i=1; i<dim; i++){
        double v = ex.x[i];
        double delta = v - mean[label][i];
        double R = delta * ex.w / sum_w;
        mean[label][i] += R;
        if(old_sum_w > 0.0){
            M2[label][i] += old_sum_w * delta * R ;
        }
    }

}

double pi2 = 2*M_PI;
double NaiveBayesGauss::raw_predict(simple_sparse_vector& x)
{
    if(sum_w < 1e-15)
        return 0.0;
    double prob[2];

    for(int c=0; c<2; ++c){
        double p = pc[c]/sum_w;
        if(p < 1e-15) {
            if(pc[1-c]/sum_w < 1e-15)
                return 0;
            else
                return 1-2*c;
        }
        p = log(p);
        double* mean_c = mean[c];
        double* M2_c = M2[c];

        uint last = 1;
        for(simple_sparse_vector_iterator it = x.my_vec.begin() + 1; 
                it != x.my_vec.end(); it++) {
            for(uint i=last; i<it->first; ++i){
                double var = (sum_w > 1.0)? M2_c[i]/(sum_w - 1.0) : 1e-16;
                double pfc = -0.5 * (log(pi2*var) + mean_c[i]*mean_c[i]/var);
                p += pfc;
            }
            uint i = it->first;
            double v = it->second;
            double var = (sum_w > 1.0)? M2_c[i]/(sum_w - 1.0) : 1e-16;
            double tmp = (v-mean_c[i]);
            double pfc = -0.5 * (log(pi2*var) + tmp*tmp/var);
            p += pfc;
            last = it->first + 1;
        }
        for(int i=last; i<dim; ++i){
            double var = (sum_w > 1.0)? M2_c[i]/(sum_w - 1.0) : 1e-16;
            double pfc = -0.5 * (log(pi2*var) + mean_c[i]*mean_c[i]/var);
            p += pfc;
        }
        prob[c] = p;
    }
    prob[1] = 1.0/(1.0+exp(prob[0]-prob[1]));
    return 2.0 * prob[1] - 1.0;
}

void NaiveBayesGauss::reset()
{
    if(mean[0]==NULL)
        mean[0] = new double[dim]();
    if(mean[1]==NULL)
        mean[1] = new double[dim]();
    if(M2[0]==NULL)
        M2[0] = new double[dim]();
    if(M2[1]==NULL)
        M2[1] = new double[dim]();

    t = 0;
    sum_w = 2.0 * 1e-16;;
    for(int c=0; c<2; ++c){
        pc[c] = 1e-16;
        for(int i=1; i<dim; ++i){
            mean[c][i] = 1e-16;
            M2[c][i] = 1e-16;
        }
    }
}

//==============================================//
//------------- Naive Bayes (Binomial) ---------//
//==============================================//
void NaiveBayesBinary::update(Example& ex)
{
    int label = (ex.y == -1)? 0 : 1;
    sum_w += ex.w;

    // update P( C )
    pc[label] += ex.w;

    // update P( Fi | C )
    for(int i=1; i<dim; i++){
        if(ex.x[i] < 1e-15)
            neg[label][i] += ex.w;
        else
            pos[label][i] += ex.w;
    }
}

double NaiveBayesBinary::raw_predict(simple_sparse_vector& x)
{
    if(sum_w < 1e-15)
        return 0.0;
    double prob[2];
    for(int c=0; c<2; ++c){
        double p = 1e+300 * pc[c]/sum_w;
        if(p < 1e-15) {
            if(pc[1-c]/sum_w < 1e-15)
                return 0;
            else
                return 1-2*c;
        }
        for(int i=1; i<dim; ++i){
            if(x[i] == 0.0){
                double tmp = neg[c][i] / pc[c];
                p *= tmp;
            }
            else{
                double tmp = pos[c][i] / pc[c];
                p *= tmp;
            }

        }
        prob[c] = p;
    }
    double sum = prob[0] + prob[1];
    if(sum>1e-15){
        return 2.0 * (prob[1]/sum) - 1.0;
    }
    else
        return 0;
}

void NaiveBayesBinary::reset()
{
    if(pos[0]==NULL)
        pos[0] = new double[dim]();
    if(pos[1]==NULL)
        pos[1] = new double[dim]();
    if(neg[0]==NULL)
        neg[0] = new double[dim]();
    if(neg[1]==NULL)
        neg[1] = new double[dim]();

    sum_w = 2.0 * 1e-16;
    for(int c=0; c<2; ++c){
        pc[c] = 1e-16;
        for(int i=1; i<dim; ++i){
            pos[c][i] = 1e-16;
            neg[c][i] = 1e-16;
        }
    }
}

//==============================================//
//---------- OSBoost (uniform weighted) --------//
//==============================================//
OSBoost::OSBoost(double g) : N(0), gamma(g), theta(g/(2+g)) { }

void OSBoost::update(Example& ex)
{
    double zt = 0.0;
    for(int i=0; i<N; i++){
        zt += weak_learner[i]->normalized_predict(ex.x) * ex.y - theta;
        weak_learner[i]->update(ex);
        ex.w = (zt<=0)? 1.0 : pow(1.0-gamma, zt/2.0);
    }
}

double OSBoost::raw_predict(simple_sparse_vector& x)
{
    double sum = 0.0;
    for(int i=0; i<N; i++){
        sum += alpha[i] * weak_learner[i]->normalized_predict(x);
    }
    return sum;
}

void OSBoost::add_weak_learner(Learner* learner)
{
    weak_learner.push_back(learner);
    N++;
}

void OSBoost::reset()
{
    alpha = new double[N]();
    for(int i=0; i<N; i++){
        alpha[i] = 1.0/N;
    }
    for(int i=0; i<N; i++){
        weak_learner[i]->reset();
        alpha[i] = 1.0/N;
    }
}

//==============================================//
//---------------- OSBoost.OCP -----------------//
//==============================================//
void OSBoostOCP::update(Example& ex)
{
    ++t;
    if(raw_predict(ex.x) * ex.y < theta) {
        double eta = 1/sqrt(t);
        for(int i=0; i<N; i++)
            alpha[i] += eta * ex.y * weak_learner[i]->normalized_predict(ex.x);
        project_to_simplex(alpha, N, 1.0);
    }
    OSBoost::update(ex);
}

//==============================================//
//----------------- OSBoost.EXP ----------------//
//==============================================//
void OSBoostEXP::update(Example& ex)
{
    double beta = 0.5; 
    double exp_predict = 0.0;
    for(int i=0; i<N; i++){
        exp_predict += weak_learner[i]->normalized_predict(ex.x);
        if(exp_predict * ex.y <= 0)
            alpha[i] *= beta;
    }
    OSBoost::update(ex);
}

double OSBoostEXP::raw_predict(simple_sparse_vector& x)
{
    int k = choose(alpha, N); 
    double sum = 0.0;
    for(int i=0; i<k; i++){
        sum += weak_learner[i]->normalized_predict(x);
    }
    return sum;
}


//==============================================//
//----------------- OzaBoost -------------------//
//==============================================//
OzaBoost::OzaBoost() : N(0)
{
    gsl_rng_env_setup();
    r = gsl_rng_alloc(gsl_rng_default);
}

OzaBoost::~OzaBoost()
{
    delete[] w_correct;
    delete[] w_wrong;
    delete r;
}

void OzaBoost::update(Example& ex)
{
    double tmp_weight = ex.w;
    double lambda = 1.0;
    for(int i=0; i<N; i++) {
        int k = gsl_ran_poisson(r, lambda);
        for(int j=0; j<k; j++)
            weak_learner[i]->update(ex);
        if(weak_learner[i]->predict(ex.x) == ex.y) {
            w_correct[i] += lambda;
            lambda *= (w_correct[i] + w_wrong[i]) / (2 * w_correct[i]);
        }else{
            w_wrong[i] += lambda;
            lambda *= (w_correct[i] + w_wrong[i]) / (2 * w_wrong[i]);
        }
    }
    ex.w = tmp_weight;
}

double OzaBoost::raw_predict(simple_sparse_vector& x)
{
    double sum = 0.0;
    for(int i=0; i<N; i++) {
        sum += log((w_correct[i]+1e-16)/(w_wrong[i]+1e-16)) * weak_learner[i]->predict(x);
    }
    return sum;
}

void OzaBoost::add_weak_learner(Learner* learner)
{
    weak_learner.push_back(learner);
    N++;
}

void OzaBoost::reset()
{
    w_correct = new double[N]();
    w_wrong = new double[N]();
    for(int i=0; i<N; i++){
        weak_learner[i]->reset();
    }
}

//==============================================//
//---------- OGBoost (uniform weighted) --------//
//==============================================//
OGBoost::OGBoost() : N(0) {}

void OGBoost::update(Example& ex)
{
    double tmp_weight = ex.w;
    double F = 0.0;
    ex.w = 0.5;
    for(int i=0; i<N; i++){
        weak_learner[i]->update(ex);
        F += weak_learner[i]->normalized_predict(ex.x);
        ex.w = 1.0/(1.0+exp(ex.y * F));
    }
    ex.w = tmp_weight;
}

double OGBoost::raw_predict(simple_sparse_vector& x)
{
    double sum = 0.0;
    for(int i=0; i<N; i++){
        sum += weak_learner[i]->normalized_predict(x);
    }
    return sum;
}

void OGBoost::add_weak_learner(Learner* learner)
{
    weak_learner.push_back(learner);
    N++;
}

void OGBoost::reset()
{
    for(int i=0; i<N; i++){
        weak_learner[i]->reset();
    }
}


