//=============================================================================
// File Name: osboost.h
// Written by: Shang-Tse Chen (27.05.12)
//=============================================================================
#include <cmath>
#include <vector>
#include <algorithm>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "stream.h"

// some tool functions
void project_to_simplex(double* x, int n, double b);
void L2_normalize(double* x, int n, double b);
int choose(double*p, int n);

// general learner class
class Learner
{
    public:
        virtual void update(Example&) = 0;
        virtual double raw_predict(simple_sparse_vector&) = 0;
        virtual double normalized_predict(simple_sparse_vector&) = 0;
        int predict(simple_sparse_vector&);
        virtual void reset() = 0;
};

class Perceptron : public Learner
{
    public:
        Perceptron(int d) : w(d), dim(d), t(0) { }; 
        void update(Example&);
        double raw_predict(simple_sparse_vector&);
        double normalized_predict(simple_sparse_vector&);
        void reset();

    private:
        WeightVector w;
        int dim;
        int t;
};

class NaiveBayesGauss : public Learner
{
    public:
        NaiveBayesGauss(int d) : t(0), dim(d), sum_w(2.0*1e-16) {
            mean[0] = NULL;
            mean[1] = NULL;
            M2[0] = NULL;
            M2[1] = NULL;
        }; 
        ~NaiveBayesGauss(){
            delete[] mean[0];
            delete[] mean[1];
            delete[] M2[0];
            delete[] M2[1];
        }
        void update(Example&);
        double raw_predict(simple_sparse_vector&);
        double normalized_predict(simple_sparse_vector& x) {return raw_predict(x);}
        void reset();

    private:
        int t;
        int dim;
        double sum_w;
        double pc[2] ;
        double* mean[2];
        double* M2[2];
};

class NaiveBayesBinary : public Learner
{
    public:
        NaiveBayesBinary(int d) : dim(d), sum_w(2.0*1e-16) {
            pos[0] = NULL;
            pos[1] = NULL;
            neg[0] = NULL;
            neg[1] = NULL;
        }; 
        ~NaiveBayesBinary(){
            delete[] pos[0];
            delete[] pos[1];
            delete[] neg[0];
            delete[] neg[1];
        }
        void update(Example&);
        double raw_predict(simple_sparse_vector&);
        double normalized_predict(simple_sparse_vector& x) {return raw_predict(x);}
        void reset();

    private:
        int dim;
        double sum_w;
        double pc[2] ;
        double* pos[2];
        double* neg[2];
};

class OSBoost : public Learner
{
    public:
        OSBoost(double g);
        ~OSBoost() {delete[] alpha;}
        void update(Example&);
        double raw_predict(simple_sparse_vector&);
        double normalized_predict(simple_sparse_vector& x) {return raw_predict(x);};
        void add_weak_learner(Learner*);
        void reset();

    protected:
        std::vector<Learner*> weak_learner;
        int N;
        double gamma;
        double theta;
        double* alpha;
};

class OSBoostOCP : public OSBoost
{
    public:
        OSBoostOCP(double g) : OSBoost(g), t(0) { }
        void update(Example&);

    private:
        int t;
};

class OSBoostEXP : public OSBoost
{
    public:
        OSBoostEXP(double g) : OSBoost(g){ }
        void update(Example&);
        double raw_predict(simple_sparse_vector&);

};

class OzaBoost : public Learner
{
    public:
        OzaBoost();
        ~OzaBoost();
        void update(Example&);
        double raw_predict(simple_sparse_vector&);
        double normalized_predict(simple_sparse_vector& x) {return raw_predict(x);};
        void add_weak_learner(Learner*);
        void reset();

    protected:
        std::vector<Learner*> weak_learner;
        int N;
        double* w_correct;
        double* w_wrong;
        gsl_rng * r;
};

class OGBoost : public Learner
{
    public:
        OGBoost();
        void update(Example&);
        double raw_predict(simple_sparse_vector&);
        double normalized_predict(simple_sparse_vector& x) {return raw_predict(x);};
        void add_weak_learner(Learner*);
        void reset();

    protected:
        std::vector<Learner*> weak_learner;
        int N;
};
